import os
import errno
import hashlib
from re import findall
from time import time
from traceback import print_exc
from requests import get, head


def download_file(url, path, filename=None, headers=None, verify=True):
    # NOTE the stream=True parameter
    if __is_downloadable(url, verify):
        r = get(url, stream=True, headers=headers,
                allow_redirects=True, verify=verify)
        if filename is None:
            filename = __get_filename_from_cd(
                r.headers.get('content-disposition'))
        file_path = "{0}/downloads/{1}".format(path, filename)
        create_local_project_export_structure(os.path.dirname(file_path))
        with open(file_path, "wb") as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
    return filename


def create_local_project_export_structure(dir_path):
    if not os.path.exists(dir_path):
        try:
            os.makedirs(dir_path)
        except OSError as exc:  # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise


def __is_downloadable(url, verify):
    """
        Does the url contain a downloadable resource
    """
    h = head(url, allow_redirects=True, verify=verify)
    header = h.headers
    content_type = header.get('content-type')
    if 'text' in content_type.lower():
        return False
    if 'html' in content_type.lower():
        return False
    return True


def __get_filename_from_cd(cd):
    """
        Get filename from content-disposition
    """
    if not cd:
        return None
    fname = findall('filename=(.+)', cd)
    if len(fname) == 0:
        return None
    return fname[0]


def is_recent_file(path, age=2592000):
    """
        Check whether a file path exists, is empty and older than 1 month
    """
    return os.path.exists(path) and os.path.getsize(
        path) > 0 and (time() - os.path.getmtime(path) < age)


def get_hash_of_dirs(directory, verbose=0):
    """
        http://akiscode.com/articles/sha-1directoryhash.shtml
        Copyright (c) 2009 Stephen Akiki
        MIT License (Means you can do whatever you want with this)
        See http://www.opensource.org/licenses/mit-license.php
        Error Codes:
        -1 -> Directory does not exist
        -2 -> General error (see stack traceback)
    """
    SHAhash = hashlib.sha1()
    if not os.path.exists(directory):
        return -1

    try:
        for root, _, files in os.walk(directory):
            for names in files:
                if verbose == 1:
                    print('Hashing', names)
                filepath = os.path.join(root, names)
                f1 = None
                try:
                    f1 = open(filepath, 'rb')
                except BaseException:
                    # You can't open the file for some reason
                    f1.close()
                    continue

                while True:
                    # Read file in as little chunks
                    buf = f1.read(4096)
                    if not buf:
                        break
                    SHAhash.update(hashlib.sha1(buf).hexdigest().encode())
                f1.close()
    except BaseException:
        # Print the stack traceback
        print_exc()
        return -2

    return SHAhash.hexdigest()


def find_files_in_folder(wildcard, directory):
    return [f for f in os.listdir(directory) if wildcard in f]



import base64
import re

from collections import Counter
from copy import deepcopy
from re import sub
from datetime import timedelta, date, datetime
from types import GeneratorType
from urllib.parse import urlparse
from requests import Response


def remove_dupes_but_take_higher_access(my_list):
    """
        Deduping function for keeping members with higher access
    """
    already_found = {}
    new_list = []
    for d in my_list:
        obj_id = d["id"]
        if already_found.get(obj_id):
            if already_found[obj_id]["access_level"] < d["access_level"]:
                c = deepcopy(d)
                c["index"] = already_found[obj_id]["index"]
                new_list[already_found[obj_id]["index"]] = c
                already_found[obj_id] = c
        else:
            already_found[obj_id] = deepcopy(d)
            new_list.append(d)
            already_found[obj_id]["index"] = len(new_list) - 1
    return new_list


def expiration_date():
    return (date.today() + timedelta(days=2)).strftime('%Y-%m-%d')


def parse_query_params(params):
    query_params_string = ""
    query_params_list = []
    for p in params:
        if params.get(p, None) is not None:
            query_params_list.append("%s=%s" % (p, str(params[p])))

    if len(query_params_list) > 0:
        query_params_string = "?%s" % "&".join(query_params_list)

    return query_params_string


def input_generator(params):
    for param in params:
        yield param


def get_dry_log(dry_run=True):
    return "DRY-RUN: " if dry_run else ""


def get_rollback_log(rollback=False):
    return "Rollback: " if rollback else ""


def pretty_print_key(s):
    return " ".join(w.capitalize() for w in s.split("_"))


def is_error_message_present(response):
    errors = ["message", "errors", "error"]
    if isinstance(response, Response):
        response = safe_json_response(response)
    if isinstance(response, (GeneratorType, map, filter)):
        response = list(response)
    if isinstance(response, list) and response and response[0] in errors:
        return True, response
    if isinstance(response, dict) and any(r in response for r in errors):
        return True, response
    if isinstance(response, str) and response in errors:
        return True, response
    return False, response


def get_timedelta(timestamp):
    """
    Get timedelta between provided timestamp and current time

        :param timestamp: A timestamp string
        :return: timedelta between provided timestamp and datetime.now() in hours
    """
    try:
        created_at = datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%SZ')
    except ValueError:
        created_at = datetime.strptime(
            timestamp.split(".")[0], '%Y-%m-%dT%H:%M:%S')
    now = datetime.now()
    return (now - created_at).days * 24


def validate_name(name, log=None):
    """
    Validate group and project names to satisfy the following criteria:
    Name can only contain letters, digits, emojis, '_', '.', dash, space.
    It must start with letter, digit, emoji or '_'.
    """
    valid = " ".join(sub(r"[^\U00010000-\U0010ffff\w\_\-\. ]",
                         " ", name.lstrip("-").lstrip(".")).split())
    if name != valid:
        output = f"Renaming invalid name '{name}' -> '{valid}'"
        log.warning(output) if log else print(output)
    return valid


def generate_audit_log_message(req_type, message, url, data=None):
    try:
        return "{0}enerating {1} request to {2}{3}".format(
            "{} by g".format(message) if message else "G",
            req_type,
            url,
            " with data: {}".format(data) if data else "")
    except TypeError as te:
        return f"Message formatting ERROR ({te}). No specific message generated. Generating {req_type} request to {url}"


def safe_json_response(response):
    """
        Helper method to handle getting valid JSON safely. If valid JSON cannot be returned, it returns none.
    """
    if response is not None:
        try:
            if isinstance(response, GeneratorType):
                return list(response)
            return response.json()
        except ValueError:
            return None
    return None


def get_duplicate_paths(data, are_projects=True):
    """
        Legacy GL versions had case insensitive paths, which on newer GL versions are seen as duplicates
    """
    paths = [x.get("path_with_namespace", "").lower() if are_projects else x.get(
        "full_path", "").lower() for x in data]
    return [i for i, c in Counter(paths).items() if c > 1]


def strip_netloc(s):
    return urlparse(s).netloc


def strip_scheme(s):
    return urlparse(s).scheme


def get_decoded_string_from_b64_response_content(response):
    """
        Takes a web response, returns the decoded *string* of the content, not byte object
    """
    if j := safe_json_response(response):
        content = j.get("content", "")
        if content is not None and str("content").strip() != "":
            return base64.b64decode(content).decode()
    return None


def do_yml_sub(yml_file, pattern, replace_with):
    """
        Does a regex subn and returns the entity
    """
    return re.subn(pattern, replace_with, yml_file)



from traceback import print_exc
from os import getenv
from time import sleep
from functools import wraps
from gitlab_ps_utils.logger import myLogger
from gitlab_ps_utils.exceptions import ConfigurationException

log = myLogger(__name__, app_path=getenv('APP_PATH', '.'),
               log_name=getenv('APP_NAME', 'application'))


def stable_retry(function, ExceptionType=Exception, delay=5, backoff=1.20):
    @wraps(function)
    def f_retry(*args, **kwargs):
        retries = 3
        mretries, mdelay = retries, delay
        while mretries >= 0:
            try:
                return function(*args, **kwargs)
            except ConfigurationException:
                return function(*args, **kwargs)
            except ExceptionType as e:
                log.error(
                    "{0}, Api connecion failed Retrying in {1} seconds...".format(e, mdelay))
                log.error(print_exc())
                sleep(mdelay)
                mretries -= 1
                mdelay *= backoff
        log.error("Failed to connect to API within {0} retr{1}".format(
            retries, "y" if retries else "ies"))
    return f_retry


def configurable_stable_retry(ExceptionType=Exception, retries=3, delay=5, backoff=1.20):
    def stable_retry(function, ExceptionType=ExceptionType,
                     retries=retries, delay=delay, backoff=backoff):
        def f_retry(*args, **kwargs):
            mretries, mdelay = retries, delay
            while mretries >= 0:
                try:
                    return function(*args, **kwargs)
                except ExceptionType as e:
                    log.error(
                        "{0}, Api connecion failed Retrying in {1} seconds...".format(e, mdelay))
                    sleep(mdelay)
                    mretries -= 1
                    mdelay *= backoff
            log.error("Failed to connect to API within {0} retr{1}".format(
                retries, "y" if retries else "ies"))
            log.error(print_exc())
        return f_retry
    return stable_retry


def token_rotate(function):
    """
        Decorator used to rotate token used from a list

        This decorator assumes args[0] is `self` in a class
        and the class needs to have a `token_array` instance
        attribtue and `index` class attribute
    """
    @wraps(function)
    def f_rotate(*args, **kwargs):
        if args[0].token_array:
            log.info("Rotating token")
            args[0].index += 1
            index = args[0].index % len(args[0].token_array)
            args[0].token = args[0].token_array[index]
        else:
            log.info("No tokens provided")
        return function(*args, **kwargs)
    return f_rotate



from logging import getLogger, FileHandler, StreamHandler, Formatter, Handler, INFO, WARNING

from os.path import exists
import json
import requests

loggers = {}
log_file_format = \
    "[%(asctime)s][%(levelname)s]|%(module)s.%(funcName)s:%(lineno)d| %(message)s"


def myLogger(name, app_path='.', log_dir='data/logs', log_name='application', config=None):
    global loggers
    if loggers.get(name):
        return loggers.get(name)
    else:
        logger = getLogger(name)
        logger.setLevel(INFO)
        stderr_log_handler = StreamHandler()
        formatter = Formatter(log_file_format, datefmt="%d %b %Y %H:%M:%S")
        log_dir = f'{app_path}/{log_dir}'
        if exists(log_dir):
            log_path = f'{log_dir}/{log_name}.log'
            file_log_handler = FileHandler(log_path)
            file_log_handler.setFormatter(formatter)
            logger.addHandler(file_log_handler)
        stderr_log_handler.setFormatter(formatter)
        logger.addHandler(stderr_log_handler)
        slack_handler = SlackLogHandler(config=config)
        slack_handler.setLevel(WARNING)
        logger.addHandler(slack_handler)
        loggers[name] = logger

        return logger


class SlackLogHandler(Handler):
    EMOJIS = {
        "CRITICAL": ":boom:",
        "ERROR": ":x:",
        "WARNING": ":warning:",
        "NOTSET": ":question:",
    }

    def __init__(self, config=None):
        Handler.__init__(self)
        self.config = config

    def emit(self, record):
        try:
            if self.config and self.config.slack_url:
                json_data = json.dumps({
                    "text": f"{record.asctime} - {record.module} - {record.msg}",
                    "username": f"{record.levelname} - {record.process}",
                    "icon_emoji": self.EMOJIS.get(record.levelname, self.EMOJIS["NOTSET"])
                })
                requests.post(self.config.slack_url, data=json_data.encode(
                    'ascii'), headers={'Content-Type': 'application/json'})
        except Exception as em:
            print("EXCEPTION: " + str(em))



import hashlib
from json import dumps as json_dumps
from gitlab_ps_utils.xml_utils import safe_xml_parse

def rewrite_list_into_dict(l, comparison_key, prefix="", lowercase=False):
    """
    Rewrites list of dictionaries into a dictionary for easier nested dict lookup

        :param: l: (list) list to convert to a dictionary
        :param: comparison_key: (str) key to use for lookup. Needs to be a unique value within the nested dictionaries like an ID
        :param: prefix: (str) optional string to use as a prefix for the key lookup
        :param: lowercase: (bool) will convert all comparison keys to lowercase to avoid any issues with case sensitive key lookups
        :return: (dict) rewritten dictionary
    """
    rewritten_obj = {}
    for i, _ in enumerate(l):
        new_obj = l[i]
        key = l[i][comparison_key]
        if prefix:
            key = prefix + str(key)
        if lowercase:
            rewritten_obj[str(key).lower()] = new_obj
        else:
            rewritten_obj[key] = new_obj

    return rewritten_obj


def rewrite_json_list_into_dict(l):
    """
        Converts a JSON list:
        [
            {
                "hello": {
                    "world": "how are you"
                }
            },
            {
                "world": {
                    "how": "are you"
                }
            }
        ]

        to:
        {
            "hello": {
                "world": "how are you"
            },
            "world": {
                "how": "are you"
            }
        }

        Note: The top level keys in the nested objects must be unique or else data will be overwritten
    """
    new_dict = {}
    for i, _ in enumerate(l):
        key = list(l[i].keys())[0]
        new_dict[key] = l[i][key]
    return new_dict

def sanitize_booleans_in_dict(d):
    """
        Helper method to convert string representations of boolean values to boolean type
    """
    for k, v in d.items():
        if isinstance(v, dict):
            sanitize_booleans_in_dict(v)
        if isinstance(v, str):
            if v.lower() == 'false':
                d[k] = False
            elif v.lower() == 'true':
                d[k] = True
    return d

def find(key, dictionary):
    """
        Nested dictionary lookup from https://gist.github.com/douglasmiranda/5127251
    """
    if isinstance(dictionary, dict):
        for k, v in dictionary.items():
            if k == key:
                yield v
            elif isinstance(v, dict):
                for result in find(key, v):
                    yield result
            elif isinstance(v, list):
                for d in v:
                    for result in find(key, d):
                        yield result


def dig(dictionary, *args, default=None):
    """
        Recursive dictionary key lookup function

        Example:
            dig({"nest": {"hello": {"world": "this is nested"}}}, "nest", "hello")
            >>>> {'world': 'this is nested'}

        :param dictionary: (dict) dictionary to traverse
        :param *args: (tuple) series of keys to dig through
        :return: If the most nested key is found, the value of the key

    """
    if not args:
        return dictionary
    if isinstance(dictionary, dict):
        for i, arg in enumerate(args):
            found = dictionary.get(arg, None)
            if found is not None:
                if isinstance(found, dict):
                    args = args[i + 1:]
                    return dig(found, *args, default=default)
                return found
            return default
    return default


def list_to_dict(lst):
    """
    Convert list to dictionary for unique key comparison
    Example input:
        [1, 2, 3, 4, 5]
    Example output:
        {
            1: True,
            2: True,
            3: True,
            4: True,
            5L True
        }

        :param lst: list to convert
        :return: dictionary converted from list
    """
    res_dct = {lst[i]: True for i in range(0, len(lst), 2)}
    return res_dct


def get_hash_of_dict(d):
    SHAhash = hashlib.sha1()
    SHAhash.update(bytes(json_dumps(d), encoding="UTF-8"))
    return SHAhash.hexdigest()


def are_keys_in_dict(list_of_keys, dictionary):
    keys_in_dict = False
    for k in list_of_keys:
        if k in dictionary.keys():
            keys_in_dict = True
    return keys_in_dict


def is_nested_dict(d):
    if isinstance(d, dict):
        return any(isinstance(i, dict) for i in d.values())
    return False


def pop_multiple_keys(src, keys):
    for k in keys:
        src.pop(k, None)
    return src


def sort_dict(d):
    """
        Sorts dictionary by key name in descending order
    """
    return {k: d[k] for k in sorted(d.keys())}


def xml_to_dict(data):
    return sanitize_booleans_in_dict(safe_xml_parse(data))



import sys
import json
from traceback import print_exc


def json_pretty(data):
    return json.dumps(data, indent=4, sort_keys=True)


def write_json_to_file(path, data, log=None):
    if log:
        log.info(f"### Writing output to {path}")
    with open(path, "w") as f:
        json.dump(data, f, indent=4)


def read_json_file_into_object(path):
    try:
        with open(path, "r") as f:
            return json.load(f)
    except FileNotFoundError:
        sys.exit(f"{path} not found")


def write_json_yield_to_file(file_path, generator_function, *args):
    with open(file_path, "w") as f:
        output = []
        for data in generator_function(*args):
            output.append(data)
        f.write(json_pretty(output))


def stream_json_yield_to_file(
        file_path, generator_function, *args, log=None, **kwargs):
    with open(file_path, 'w') as f:
        f.write("[\n")
        try:
            for data, last_result in generator_function(*args, **kwargs):
                f.write(json_pretty(data))
                if not last_result:
                    f.write(",")
        except Exception as e:
            if log:
                log.error("Streamed write failed with error:\n{}".format(e))
                log.error(print_exc())
            else:
                print_exc()
        finally:
            f.write("\n]")



import sys
from os import getenv

from traceback import print_exc
from multiprocessing import Pool, cpu_count, get_context
from functools import partial
from tqdm import tqdm
from gitlab_ps_utils.json_utils import json_pretty
from gitlab_ps_utils.logger import myLogger
from gitlab_ps_utils.process import NoDaemonProcess


class MultiProcessing():
    tanuki = "#e24329"
    _func = None
    def __init__(self):
        self.log = myLogger(__name__, app_path=getenv('APP_PATH', '.'), log_name=getenv('APP_NAME', 'application'))

    def worker_init(self, func):
        global _func
        _func = func


    def worker(self, x):
        return _func(x)


    def start_multi_process(self, function, iterable, processes=None, nestable=False):
        """
            Wrapper function to handle multiprocessing a function with a list of data

            This function leverages map to handle multiprocessing. This function will return a list of data from the function so use this function if your function returns necessary data

            :param: function: (func) The function processesing the elements of the list
            :param: iterable: (list) A list of data to be passed into the function to process
            :param: processes: (int) Explicit number of processes to split the function across. If processes is not set, number of processes will default to total number of physical cores of CPU
            :param: nestable: (bool) Allow this multiprocessed function to create nested multiprocessed functions

            :return: A list of the data returned from the process map
        """
        ctx = get_context("spawn")
        if nestable:
            ctx.Process = NoDaemonProcess
        p = ctx.Pool(processes=self.get_no_of_processes(processes),
                    initializer=self.worker_init, initargs=(function,))
        try:
            for i in tqdm(p.imap_unordered(self.worker, iterable),
                        total=len(iterable), colour=self.tanuki):
                yield i
        except Exception as e:
            self.log.critical("Migration pool failed with error:\n{}".format(e))
            self.log.critical(print_exc())
        finally:
            p.close()
            p.join()


    def start_multi_process_with_args(
            self, function, iterable, *args, processes=None, nestable=False, **kwargs):
        """
            Wrapper function to handle multiprocessing a function with multiple arguments with a list of data

            example function signature:

            def func(*args, iterable, **kwargs)

            The iterable must be the last argument in the handler function signature

            This function leverages map to handle multiprocessing. This function will return a list of data from the function so use this function if your function returns necessary data

            :param: function: (func) The function processesing the elements of the list
            :param: iterable: (list) A list of data to be passed into the function to process
            :param: *args: (args) Any additional arguments the function needs passed in
            :params processes: (int) Explicit number of processes to split the function across. If processes is not set, number of processes will default to total number of physical cores of CPU
            :param: nestable: (bool) Allow this multiprocessed function to create nested multiprocessed functions

            :return: A list of the data returned from the process map
        """
        ctx = get_context("spawn")
        if nestable:
            ctx.Process = NoDaemonProcess
        p = ctx.Pool(processes=self.get_no_of_processes(processes),
                    initializer=self.worker_init, initargs=(partial(function, *args, **kwargs),))
        try:
            for i in tqdm(p.imap_unordered(self.worker, iterable),
                        total=len(iterable), colour=self.tanuki):
                yield i
        except Exception as e:
            self.log.critical("Migration pool failed with error:\n{}".format(e))
            self.log.critical(print_exc())
        finally:
            p.close()
            p.join()


    def start_multi_process_stream(self, function, iterable, processes=None, nestable=False):
        """
            Wrapper function to handle multiprocessing a function with a list of data

            This function leverages imap_unordered to handle processing a stream of data of unknown length, like from a generator

            :param: function: (func) The function processesing the elements of the list
            :param: iterable: (list) A list of data to be passed into the function to process
            :param: processes: (int) Explicit number of processes to split the function across. If processes is not set, number of processes will default to total number of physical cores of CPU
            :param: nestable: (bool) Allow this multiprocessed function to create nested multiprocessed functions

            :return: An imap_unordered object. Assume no useful data will be returned with this function
        """
        ctx = get_context("spawn")
        if nestable:
            ctx.Process = NoDaemonProcess
        p = ctx.Pool(processes=self.get_no_of_processes(processes),
                    initializer=self.worker_init, initargs=(function,))
        try:
            return p.imap_unordered(self.worker, iterable)
        except Exception as e:
            self.log.critical("Migration pool failed with error:\n{}".format(e))
            self.log.critical(print_exc())
        finally:
            p.close()
            p.join()


    def start_multi_process_stream_with_args(
            self, function, iterable, *args, processes=None, nestable=False, **kwargs):
        """
            Wrapper function to handle multiprocessing a function with multiple arguments with a list of data

            example function signature:

            def func(*args, iterable, **kwargs)

            The iterable must be the last argument in the handler function signature

            This function leverages imap_unordered to handle processing a stream of data of unknown length, like from a generator

            :param: function: (func) The function processesing the elements of the list
            :param: iterable: (list) A list of data to be passed into the function to process
            :param: *args: (args) Any additional arguments the function needs passed in
            :param: processes: (int) Explicit number of processes to split the function across. If processes is not set, number of processes will default to total number of physical cores of CPU
            :param: nestable: (bool) Allow this multiprocessed function to create nested multiprocessed functions

            :return: An imap_unordered object. Assume no useful data will be returned with this function
        """
        ctx = get_context("spawn")
        if nestable:
            ctx.Process = NoDaemonProcess
        p = ctx.Pool(processes=self.get_no_of_processes(processes),
                    initializer=self.worker_init, initargs=(partial(function, *args, **kwargs),))
        try:
            return p.imap_unordered(self.worker, iterable)
        except Exception as e:
            self.log.critical("Migration pool failed with error:\n{}".format(e))
            self.log.critical(print_exc())
        finally:
            p.close()
            p.join()


    def handle_multi_process_write_to_file_and_return_results(
            self, function, results_function, iterable, path, processes=None):
        with open(path, 'w') as f:
            f.write("[\n")
            try:
                p = Pool(processes=self.get_no_of_processes(processes),
                        initializer=self.worker_init, initargs=(function,))
                for result in tqdm(p.imap_unordered(
                        self.worker, iterable), total=len(iterable), colour=self.tanuki):
                    f.write(json_pretty(result))
                    yield results_function(result)
            except TypeError as te:
                print_exc()
                print("Found None ({}). Stopping write to file".format(te))
            except Exception as e:
                self.log.critical(
                    "Migration processes failed with error:\n{}".format(e))
                self.log.critical(print_exc())
            else:
                f.write("\n]")
            finally:
                p.close()
                p.join()


    def get_no_of_processes(self, processes):
        try:
            proc = int(processes) if processes else 4
            self.log.info(
                f"Running command with {proc} parallel processes on {cpu_count()} CPU")
            return proc
        except ValueError:
            self.log.error(f"Input for # of processes is not an integer: {processes}")
            sys.exit()



"""
Congregate - GitLab instance migration utility

Copyright (c) 2021 - GitLab
"""

import os
import json
from re import sub, split
from configparser import ConfigParser, ParsingError
from gitlab_ps_utils.string_utils import deobfuscate


class BaseConfig(object):
    def __init__(self, path=None):
        config_path = path
        self.config = ConfigParser()
        if not os.path.exists(config_path):
            print("WARNING: No configuration found. Configuring empty file {}".format(
                config_path))
            with open(config_path, "w") as f:
                self.config.write(f)
        try:
            self.config.read(config_path)
        except ParsingError as pe:
            print("Failed to parse configuration, with error:\n{}".format(pe))
            raise SystemExit() from pe

    def option_exists(self, section, option):
        return self.config.has_option(
            section, option) and self.config.get(section, option)

    def prop(self, section, option, default=None, obfuscated=False):
        if self.option_exists(section, option):
            if not obfuscated:
                return self.config.get(section, option)
            return deobfuscate(self.config.get(section, option))
        return default

    def prop_lower(self, section, option, default=None):
        if self.option_exists(section, option):
            return self.config.get(section, option).lower()
        return default

    def prop_int(self, section, option, default=None):
        """
            Returns configuration property string casted to an int
        """
        if self.option_exists(section, option):
            return self.config.getint(section, option)
        return default

    def prop_bool(self, section, option, default=None):
        """
            Returns configuration property string casted to a bool
        """
        if self.option_exists(section, option):
            return self.config.getboolean(section, option)
        return default

    def prop_list(self, section, option, default=None):
        """
            Returns configuration property string as a list.

            For example, a configuration property stored as '[hello, world, how, are you]'
            will be returned as ["hello", "world", "how", "are", "you"]
        """
        if self.option_exists(section, option):
            return split(r', |,', sub(r'\[|\]|\"\"\"', '', self.config.get(section, option)))
        return default

    def prop_dict(self, section, option, default=None):
        """
            Returns configuration property JSON string as a dictionary
        """
        if self.option_exists(section, option):
            return json.loads(self.config.get(section, option))
        return default

    def as_obj(self):
        """
        Return entire config object (setter)
        """
        return self.config

    def as_dict(self):
        """
        Return entire config as dictionary (copy)
        """
        d = dict(self.config._sections)
        for k in d:
            d[k] = dict(self.config._defaults, **d[k])
            d[k].pop('__name__', None)
        return d





def safe_list_index_lookup(l, v):
    """
        Helper method to safely lookup the index of a list based on a specific value
    """
    return l.index(v) if v in l else None


def remove_dupes(my_list):
    """
        Basic deduping function to remove any duplicates from a list
    """
    return list({v.get("id"): v for v in my_list}.values())


def remove_dupes_with_keys(my_list, list_of_keys):
    """
        Deduping function to remove any duplicates from a list based on a set of keys
    """
    d = {}
    for v in my_list:
        key = ""
        for k in list_of_keys:
            key += str(v.get(k, ""))
        d[key] = v
    return list(d.values())



import logging
from os.path import exists

loggers = {}
log_file_format = \
    "[%(asctime)s][%(levelname)s]| %(message)s"


def audit_logger(name, app_path='.', log_dir='data/logs'):
    name = name + "_audit"
    global loggers
    if loggers.get(name):
        return loggers.get(name)
    else:
        logger = logging.getLogger(name)
        logger.setLevel(logging.INFO)
        stderr_log_handler = logging.StreamHandler()
        formatter = logging.Formatter(
            log_file_format, datefmt="%d %b %Y %H:%M:%S")
        log_dir = f'{app_path}/{log_dir}'
        if exists(log_dir):
            log_path = f'{log_dir}/audit.log'
            file_log_handler = logging.FileHandler(log_path)
            file_log_handler.setFormatter(formatter)
            logger.addHandler(file_log_handler)
        stderr_log_handler.setFormatter(formatter)
        logger.addHandler(stderr_log_handler)
        loggers[name] = logger

        return logger



from multiprocessing import context

class NoDaemonProcess(context.SpawnProcess):
    # make 'daemon' attribute always return False
    # needed for multiprocessing child processes
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gitlab_ps_utils']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.25.1,<3.0.0', 'tqdm>=4.61.0,<5.0.0', 'xmltodict>=0.12.0,<0.13.0']

setup_kwargs = {
    'name': 'gitlab-ps-utils',
    'version': '0.3.0',
    'description': 'Shared python utilities used by GitLab Professional Services tooling',
    'long_description': '# GitLab PS Utils\n\nGitLab PS utils is the foundational API and utilities libraries used by GitLab Professional Services. \nTo see the source code, project backlog and contributing guide, [check here](https://gitlab.com/gitlab-org/professional-services-automation/gitlab-ps-utils)\n\n## Install\n\n```bash\npip install gitlab-ps-utils\n```\n\n## Usage\n\nThis library contains various utility modules and classes.\nRefer to the repository source code to see available utility functions.\n\n### Importing a utility function\n\n```python\nfrom gitlab_ps_utils.string_utils import strip_numbers\n\ntest_var = "abc123"\nprint(strip_numbers(test_var))\n```\n\n### Importing a utility class\n\n```python\nfrom gitlab_ps_utils.api import GitLabApi\n\ngl_api = GitLabApi()\n\ngl_api.generate_get_request("http://gitlab.example.com", "token", "/projects")\n```\n\n## Other resources\n\n### Python-GitLab\n\nWe include a basic GitLab API wrapper class in this library. We will include specific API wrapper functions in the future.\nOur wrapper class and specific wrappers were created when [python-gitlab](https://python-gitlab.readthedocs.io/en/stable/) was in a much earlier state,\nso we continued to use our wrapper instead of switching to python-gitlab\n\nFor generic requests to the GitLab API, our wrapper is more lightweight than python-gitlab,\nbut python-gitlab is a great option for more complex API scripts\n\n',
    'author': 'GitLab Professional Services',
    'author_email': 'proserv@gitlab.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/gitlab-org/professional-services-automation/gitlab-ps-utils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



"""
JsonDiff: Library for JSON DIFF - based on opslib jsondiff
-------------------------------

+------------------------+-------------+
| This is the JsonDiff common library. |
+------------------------+-------------+
"""

try:
    import json
except ImportError:
    import simplejson as json

# from opslib.icsexception import IcsException

import logging
log = logging.getLogger(__name__)


def is_scalar(value):
    """
    Primitive version, relying on the fact that JSON cannot
    contain any more complicated data structures.
    """
    return not isinstance(value, (list, tuple, dict))


class Comparator(object):

    """
    Main workhorse for JSON Comparator
    """

    def __init__(self, fp1=None, fp2=None,
                 include=[], exclude=[], ignore_add=False):
        """
        :type fp1: object
        :param fp1: file object (opened with read permission)

        :type fp2: object
        :param fp2: file object (opened with read permission)

        :type include: list
        :param include: a list of attributes to include in the comparison

        :type exclude: list
        :param exclude: a list of attributes to exclude in the comparison

        :type ignore_add: bool
        :param ignore_add: whether to ignore the added items in the comparison

        **Example:**

        >>> from opslib.icsutils.jsondiff import Comparator
        >>> import json
        >>> old_json = {
        ...     "name": "opslib",
        ...     "version": "1.2.0",
        ...     "members": {
        ...         "role": "ops",
        ...         "group": [ "ops", "devops" ]
        ...     }
        ... }
        >>> new_json = {
        ...     "name": "opslib",
        ...     "version": "1.3.0",
        ...     "members": {
        ...         "role": "devops",
        ...         "group": [ "devops" ]
        ...     }
        ... }
        >>> json.dump(old_json, open("old.json", "w"))
        >>> json.dump(new_json, open("new.json", "w"))
        >>> fp_old = open("old.json", "r")
        >>> fp_new = open("new.json", "r")
        >>> engine = Comparator(fp_old, fp_new)
        >>> res = engine.compare_dicts()
        >>> print(json.dumps(res, sort_keys=True, indent=4))
        {
            "members": {
                "group": {
                    "0": {
                        "+++": "devops",
                        "---": "ops"
                    },
                    "1": {
                        "---": "devops"
                    }
                },
                "role": {
                    "+++": "devops",
                    "---": "ops"
                }
            },
            "version": {
                "+++": "1.3.0",
                "---": "1.2.0"
            }
        }

        """
        self.obj1 = None
        self.obj2 = None
        if fp1:
            try:
                self.obj1 = json.load(fp1)
            except (TypeError, OverflowError, ValueError) as exc:
                raise json.JSONDecodeError("Cannot decode object from JSON.\n%s" %
                                           str(exc))
        if fp2:
            try:
                self.obj2 = json.load(fp2)
            except (TypeError, OverflowError, ValueError) as exc:
                raise json.JSONDecodeError("Cannot decode object from JSON\n%s" %
                                           str(exc))

        self.excluded_attributes = []
        self.included_attributes = []
        self.ignore_added = False
        if include:
            self.included_attributes = include or []
        if exclude:
            self.excluded_attributes = exclude or []
        if ignore_add:
            self.ignore_added = ignore_add or False

    def _is_incex_key(self, key, value):
        """Is this key excluded or not among included ones? If yes, it should
        be ignored."""
        key_out = ((self.included_attributes and
                    (key not in self.included_attributes)) or
                   (key in self.excluded_attributes))
        value_out = True
        if isinstance(value, dict):
            for change_key in value:
                if isinstance(value[change_key], dict):
                    for key in value[change_key]:
                        if ((self.included_attributes and
                             (key in self.included_attributes)) or
                                (key not in self.excluded_attributes)):
                            value_out = False
        return key_out and value_out

    def _filter_results(self, result):
        """Whole -i or -x functionality. Rather than complicate logic while
        going through the object's tree we filter the result of plain
        comparison.

        Also clear out unused keys in result"""
        out_result = {}
        for change_type in result:
            temp_dict = {}
            for key in result[change_type]:
                log.debug("change_type = %s", change_type)
                if self.ignore_added and (change_type == "+++"):
                    continue
                log.debug("result[change_type] = %s, key = %s",
                          str(result[change_type]), key)
                log.debug("self._is_incex_key = %s",
                          self._is_incex_key(
                              key,
                              result[change_type][key]))
                if not self._is_incex_key(key, result[change_type][key]):
                    temp_dict[key] = result[change_type][key]
            if len(temp_dict) > 0:
                out_result[change_type] = temp_dict

        return out_result

    def _compare_elements(self, old, new):
        """Unify decision making on the leaf node level."""
        res = None
        # We want to go through the tree post-order
        if isinstance(old, dict):
            res_dict = self.compare_dicts(old, new)
            if (len(res_dict) > 0):
                res = res_dict
        # Now we are on the same level
        # different types, new value is new
        elif (type(old) != type(new)):
            res = {'---': old, '+++': new}
        # recursive arrays
        # we can be sure now, that both new and old are
        # of the same type
        elif (isinstance(old, list)):
            res_arr = self._compare_arrays(old, new)
            if (len(res_arr) > 0):
                res = res_arr
        # the only thing remaining are scalars
        else:
            scalar_diff = self._compare_scalars(old, new)
            if scalar_diff is not None:
                res = scalar_diff

        return res

    def _compare_scalars(self, old, new, name=None):
        """
        Be careful with the result of this function. Negative answer from this
        function is really None, not False, so deciding based on the return
        value like in

        if self._compare_scalars(...):

        leads to wrong answer (it should be
        if self._compare_scalars(...) is not None:)
        """
        # Explicitly excluded arguments
        if old != new:
            return {'---': old, '+++': new}
        else:
            return None

    def _compare_arrays(self, old_arr, new_arr):
        """
        simpler version of compare_dicts; just an internal method, because
        it could never be called from outside.

        We have it guaranteed that both new_arr and old_arr are of type list.
        """
        old_arr = sorted(old_arr, key=lambda x: str(x))
        new_arr = sorted(new_arr, key=lambda x: str(x))
        inters = min(len(old_arr), len(new_arr))  # this is the smaller length

        result = {
            u"+++": {},
            u"---": {},
        }

        # old_arr_dict = list_to_dict(old_arr)
        # new_arr_dict = list_to_dict(new_arr)

        for idx in range(len(old_arr)):
            if old_arr[idx] not in new_arr:
                result[idx] = {u'---': old_arr[idx]}
        for idx in range(len(new_arr)):
            if new_arr[idx] not in old_arr:
                result[idx] = {u'+++':  new_arr[idx]}
            # res = self._compare_elements(old_arr[idx], new_arr[idx])
            # if res is not None:
            #     result[idx] = res

        # the rest of the larger array
        # if (inters == len(old_arr)):
        #     for idx in range(inters, len(new_arr)):
        #         result[idx] = {u'+++':  new_arr[idx]}
        # else:
        #     for idx in range(inters, len(old_arr)):
        #         result[idx] = {u'---': old_arr[idx]}

        # Clear out unused keys in result
        out_result = {}
        for key in result:
            if len(result[key]) > 0:
                out_result[key] = result[key]

        # return self.compare_dicts(old_arr_dict, new_arr_dict)

        return self._filter_results(result)

    def compare_dicts(self, old_obj=None, new_obj=None):
        """
        The real workhorse
        """
        if old_obj is None and hasattr(self, "obj1"):
            old_obj = self.obj1
        if new_obj is None and hasattr(self, "obj2"):
            new_obj = self.obj2

        old_keys = set()
        new_keys = set()
        if old_obj and len(old_obj) > 0:
            old_keys = set(old_obj.keys())
        if new_obj and len(new_obj) > 0:
            new_keys = set(new_obj.keys())

        keys = old_keys | new_keys

        result = {
            u"+++": {},
            u"---": {},
        }

        if old_obj and new_obj:
            for name in keys:
                # old_obj is missing
                if name not in old_obj:
                    result[u'+++'][name] = new_obj[name]
                # new_obj is missing
                elif name not in new_obj:
                    result[u'---'][name] = old_obj[name]
                else:
                    res = self._compare_elements(old_obj[name], new_obj[name])
                    if res is not None:
                        result[name] = res

        return self._filter_results(result)

# vim: tabstop=4 shiftwidth=4 softtabstop=4



from math import ceil as math_ceil
from time import sleep
from re import findall
import requests

from gitlab_ps_utils.logger import myLogger
from gitlab_ps_utils.audit_logger import audit_logger
from gitlab_ps_utils.decorators import stable_retry
from gitlab_ps_utils.misc_utils import generate_audit_log_message, safe_json_response


class GitLabApi(object):
    def __init__(self, app_path=None, log_name=None, ssl_verify=True):
        self.log = myLogger(__name__, app_path=app_path, log_name=log_name)
        self.audit = audit_logger(__name__, app_path=app_path)
        self.ssl_verify = ssl_verify

    def generate_v4_request_url(self, host, api):
        return f"{host}/api/v4/{api}"

    def generate_graphql_request_url(self, host):
        return f"{host}/api/graphql"

    def generate_v4_request_header(self, token):
        return {
            'Private-Token': token,
            'Content-Type': 'application/json'
        }

    @stable_retry
    def generate_get_request(self, host, token, api, url=None, params=None):
        """
        Generates GET request to GitLab API.
        You will need to provide the GL host, access token, and specific api url.

            :param host: (str) GitLab host URL
            :param token: (str) Access token to GitLab instance
            :param api: (str) Specific GitLab API endpoint (ex: projects)
            :param url: (str) A URL to a location not part of the GitLab API. Defaults to None
            :param params:
            :return: The response object *not* the json() or text()

        """

        if url is None:
            url = self.generate_v4_request_url(host, api)

        headers = self.generate_v4_request_header(token)

        if params is None:
            params = {}

        return requests.get(url, params=params, headers=headers, verify=self.ssl_verify)

    @stable_retry
    def generate_post_request(self, host, token, api, data, url=None, graphql_query=False, headers=None, files=None, description=None):
        """
            Generates POST request to GitLab API.

            :param host: (str) GitLab host URL
            :param token: (str) Access token to GitLab instance
            :param api: (str) Specific GitLab API endpoint (ex: projects)
            :param data: (dict) Any data required for the API request

            :kwarg graphql_query: (bool) Sets the URL to use the GraphQL endpoint. Default False
            :kwarg headers: (dict) Any headers to be passed into the request. Default None
            :kwarg files: (dict) Any file content to be passed into the request. Default None
            :kwarg description: (str) A custom description message for the audit self.log. Default None

            :return: request object containing response
        """
        if graphql_query and not url:
            url = self.generate_graphql_request_url(host)
        elif not url:
            url = self.generate_v4_request_url(host, api)
        self.audit.info(generate_audit_log_message("POST", description, url))
        if headers is None:
            headers = self.generate_v4_request_header(token)
        return requests.post(url, data=data, headers=headers, files=files, verify=self.ssl_verify)

    @stable_retry
    def generate_put_request(self, host, token, api, data, headers=None, files=None, description=None):
        """
            Generates PUT request to GitLab API.

            :param host: (str) GitLab host URL
            :param token: (str) Access token to GitLab instance
            :param api: (str) Specific GitLab API endpoint (ex: projects)
            :param data: (dict) Any data required for the API request

            :return: request object containing response
        """
        url = self.generate_v4_request_url(host, api)
        self.audit.info(generate_audit_log_message("PUT", description, url))
        if headers is None:
            headers = self.generate_v4_request_header(token)

        return requests.put(url, headers=headers, data=data, files=files, verify=self.ssl_verify)

    @stable_retry
    def generate_delete_request(self, host, token, api, description=None):
        """
            Generates DELETE request to GitLab API.

            :param host: (str) GitLab host URL
            :param token: (str) Access token to GitLab instance
            :param api: (str) Specific GitLab API endpoint (ex: user/1234)

            :return: request object containing response
        """
        url = self.generate_v4_request_url(host, api)
        self.audit.info(generate_audit_log_message("DELETE", description, url))
        headers = self.generate_v4_request_header(token)

        return requests.delete(url, headers=headers, verify=self.ssl_verify)

    @stable_retry
    def get_count(self, host, token, api):
        """
            Retrieves total count of projects, users, and groups

            :param host: (str) GitLab host URL
            :param token: (str) Access token to GitLab instance
            :param api: (str) Specific GitLab API endpoint (ex: users)

            :return: long containing the data from the 'X-Total' header in the response OR None if the header doesn't exist in the response
        """
        url = self.generate_v4_request_url(host, api)

        response = requests.head(
            url, headers=self.generate_v4_request_header(token), verify=self.ssl_verify)

        if response.headers.get('X-Total', None) is not None:
            return int(response.headers['X-Total'])

    @stable_retry
    def get_total_pages(self, host, token, api):
        """
            Get total number of pages in API result

            :param host: (str) GitLab host URL
            :param token: (str) Access token to GitLab instance
            :param api: (str) Specific GitLab API endpoint (ex: users)

            :return: long containing the data from the 'X-Total-Pages' header in the response OR None if the header doesn't exist in the response
        """
        url = self.generate_v4_request_url(host, api)

        response = requests.head(
            url, headers=self.generate_v4_request_header(token), verify=self.ssl_verify)
        if response.headers.get('X-Total-Pages', None) is not None:
            return int(response.headers['X-Total-Pages'])

        return None

    @stable_retry
    def get_total_count(self, host, token, api, params=None, per_page=100, keyset=False, bypass_x_total_count=False):
        count = self.get_count(host, token, api)
        # Can't use a walrus operator if count is 0
        if (count or count == 0) and bypass_x_total_count is False:
            self.log.info(f"Total count for endpoint {api}: {count}")
            return count

        PER_PAGE = per_page
        start_at = 1
        last_id = 0

        start_page = (start_at / PER_PAGE) + 1  # pages are 1-indexed
        current_page = int(start_page)
        count = 0
        while True:
            url = self.generate_v4_request_url(host, api)
            response = requests.get(url, headers=self.generate_v4_request_header(token), verify=self.ssl_verify, params=self.get_params(
                params, PER_PAGE, current_page, keyset, last_id))
            headers = response.headers
            if headers.get('x-per-page', None):
                if data := safe_json_response(response):
                    # The system key denotes anything autogenerated by GitLab.
                    # If we include this key in our counts, it will lead to inaccurate note counts
                    if 'system' not in data[0].keys():
                        x_per_page = int(headers.get('x-per-page'))
                        self.log.info(
                            f"Retrieved {PER_PAGE * (current_page - 1) + x_per_page} {api}")
                        if keyset:
                            last_id = self.get_last_id(
                                headers.get("Link", None))
                            if last_id is None:
                                break
                        if not headers.get("x-next-page"):
                            count += len(data)
                            break
                        count += x_per_page
                    else:
                        adata = [actual_data for actual_data in data if actual_data.get(
                            "system") is False]
                        if not headers.get("x-next-page"):
                            count += len(adata)
                            break
                        count += len(adata)
            else:
                if data := safe_json_response(self.generate_get_request(host, token, api, params=self.get_params(
                        params, PER_PAGE, current_page, keyset, last_id))):
                    count += len(data)
                break
            current_page += 1
        self.log.info(f"Total count for endpoint {api}: {count}")
        return count

    @stable_retry
    def get_nested_total_count(self, host, token, apis, bypass_x_total_count=False):
        count = 0
        for top_level_data in self.list_all(host, token, apis[0]):
            if nested_id := top_level_data.get("iid", None):
                count += self.get_total_count(
                    host, token, f"{apis[0]}/{nested_id}/{apis[1]}", bypass_x_total_count=bypass_x_total_count)
        return count

    @stable_retry
    def list_all(self, host, token, api, params=None, per_page=100, keyset=False):
        """
            Generates a list of all projects, groups, users, etc.

            :param host: (str) GitLab host URL
            :param token: (str) Access token to GitLab instance
            :param api: (str) Specific GitLab API endpoint (ex: users)
            :param api: (str) Specific GitLab API endpoint (ex: users)
            :param per_page: (int) Total results per request. Defaults to 100

            :yields: Individual objects from the presumed array of data
        """

        count = self.get_count(host, token, api)
        self.log.info(f"Total count for {host} endpoint {api}: {count}")

        PER_PAGE = per_page
        start_at = 0
        end_at = count
        last_id = 0

        if count is not None:
            # total_work = end_at - start_at
            # total_pages = total_work / PER_PAGE
            start_page = (start_at / PER_PAGE) + 1  # pages are 1-indexed
            end_page = int(math_ceil(float(end_at) / float(PER_PAGE)))
            current_page = int(start_page)
            retried = False
            while current_page <= end_page:
                data = self.generate_get_request(host, token, api, params=self.get_params(
                    params, PER_PAGE, current_page, keyset, last_id))
                try:
                    self.log.info(
                        f"Retrieved {PER_PAGE * (current_page - 1) + len(data.json())} {api}")
                    if keyset:
                        last_id = self.get_last_id(
                            data.headers.get("Link", None))
                        if last_id is None:
                            break
                    data = data.json()
                    for project in data:
                        yield project
                    if len(data) < PER_PAGE:
                        break
                    current_page += 1
                    retried = False
                except ValueError as e:
                    self.log.error(e)
                    self.log.error("API request didn't return JSON")
                    self.log.info("Attempting to retry after 10 seconds")
                    sleep(10)
                    # Failure will only be retried once
                    retried = True
                if retried:
                    break
        else:
            start_page = (start_at / PER_PAGE) + 1  # pages are 1-indexed
            current_page = int(start_page)
            while True:
                data = self.generate_get_request(
                    host, token, api, params=self.get_params(
                        params, PER_PAGE, current_page, keyset, last_id))
                try:
                    self.log.info(
                        f"Retrieved {PER_PAGE * (current_page - 1) + len(data.json())} {api}")
                    if keyset:
                        last_id = self.get_last_id(
                            data.headers.get("Link", None))
                        if last_id is None:
                            break
                    data = data.json()
                    for project in data:
                        yield project
                    if len(data) < PER_PAGE:
                        break
                    current_page += 1
                except ValueError as e:
                    self.log.error(e)
                    self.log.error("API Request didn't return JSON")
                    # Retry interval is smaller here because it will just retry
                    # until it succeeds
                    self.log.info("Attempting to retry after 3 seconds")
                    sleep(3)

    @stable_retry
    def search(self, host, token, api, search_query):
        """
            Get total number of pages in API result

            :param host: (str) GitLab host URL
            :param token: (str) Access token to GitLab instance
            :param api: (str) Specific GitLab API endpoint (ex: users)
            :search_query: (str) Specific query to search

            :return: JSON object containing the request response
        """
        return self.generate_get_request(host, token, api, params={
            'search': search_query}).json()

    def get_params(self, params, per_page, current_page, keyset, last_id):
        if params:
            params["page"] = current_page
            params["per_page"] = per_page
        else:
            params = {
                "pagination": "keyset",
                "per_page": per_page,
                "order_by": "id",
                "sort": "asc",
                "id_after": last_id
            } if keyset else {
                "page": current_page,
                "per_page": per_page
            }
        return params

    # Project only keyset-based pagination - https://docs.gitlab.com/ee/api/#keyset-based-pagination

    def get_last_id(self, link):
        # Get id_after value. If the Link key is missing it's done, with an empty list response
        return findall(r"id_after=(.+?)&", link)[0] if link else None



from base64 import b64decode, b64encode
import getpass
import sys
from re import sub


def strip_numbers(s):
    """
        Strip out any numbers from a string

        :param s: (str) the string containing numbers
        :return: (str) string without numbers
    """
    return sub(r"[0-9]+", '', s)


def convert_to_underscores(s):
    """
        Converts ' ', '/', '.', ':' to underscores

        :param s: (str) the string containing spaces, slashes, periods, or colons
        :return: (str) string without those characters
    """
    return sub(r" |\/|\.|\:", "_", s)


def clean_split(s, *args, **kwargs):
    """
        Returns split string without any empty string elements

        :param: s: (str) the string to split
        :param: *args, **kwargs: any arguments you need to pass to the split function

        example usage:

        s = "hello/world"
        clean_split(s, "/")
        ['hello', 'world']
    """
    return list(filter(None, s.split(*args, **kwargs)))


def obfuscate(prompt):
    return b64encode(getpass.getpass(prompt).encode("ascii")).decode("ascii")


def deobfuscate(secret):
    try:
        return b64decode(secret.encode("ascii")).decode("ascii")
    except Exception as e:
        print(f"Invalid token - {e}")
        sys.exit()
